import asyncio
from pathlib import Path
from typing import Any
from unittest.mock import AsyncMock, MagicMock

import pytest

from torrra._types import Indexer
from torrra.app import TorrraApp
from torrra.core import config as config_module
from torrra.core.config import Config


@pytest.fixture
def fast_sleep(monkeypatch: pytest.MonkeyPatch):
    # patch asyncio sleep to almost wake up instantly
    async def no_sleep(_duration: int):
        return None

    monkeypatch.setattr(asyncio, "sleep", no_sleep)


@pytest.fixture
def app_factory():
    # returns a function to create app
    # w/ or w/o search_query
    # passing a search_query will show search screen instead.
    def _create_app(search_query: str | None = None):
        return TorrraApp(
            indexer=Indexer(
                name="jackett", url="http://mock.indexer.url", api_key="mock_api_key"
            ),
            use_cache=False,
            search_query=search_query,
        )

    return _create_app


@pytest.fixture
def mock_indexer(monkeypatch: pytest.MonkeyPatch):
    mock_indexer_instance = MagicMock()
    mock_indexer_instance.search = AsyncMock(return_value=[])

    # patch the method that creates the indexer to return mock instance
    def _mock_get_indexer_instance(self: Any):  # pyright: ignore[reportUnusedParameter]
        return mock_indexer_instance

    monkeypatch.setattr(
        "torrra.widgets.search.SearchContent._get_indexer_instance",
        _mock_get_indexer_instance,
    )
    # return patched indexer for test cases
    return mock_indexer_instance


@pytest.fixture
def mock_config(tmp_path: Path, monkeypatch: pytest.MonkeyPatch):
    # fixture to create a Config instance that uses a temp dir
    temp_config_dir = tmp_path / "torrra"
    temp_config_file = temp_config_dir / "config.toml"

    # monkeypatch constants in the config module
    monkeypatch.setattr(config_module, "CONFIG_DIR", temp_config_dir)
    monkeypatch.setattr(config_module, "CONFIG_FILE", temp_config_file)

    # the lru_cache on get_config needs to be cleared so that it doesn't
    # return a cached instance that was created before our patch was applied.
    config_module.get_config.cache_clear()

    # this will now create a Config instance using the tmp_path
    return Config()



import httpx
import respx

from torrra.utils.magnet import resolve_magnet_uri


async def test_resolve_already_magnet():
    # tests that a magnet uri is returned as is
    magnet = "magnet:?xt=urn:btih:test"
    resolved = await resolve_magnet_uri(magnet)
    assert resolved == magnet


@respx.mock
async def test_resolve_http_redirect():
    # tests that a 301/302 redirect is handled correctly
    test_url = "http://test.com/redirect"
    magnet_uri = "magnet:?xt=urn:btih:redirect"
    respx.get(test_url).mock(httpx.Response(302, headers={"location": magnet_uri}))

    resolved = await resolve_magnet_uri(test_url)
    assert resolved == magnet_uri


@respx.mock
async def test_resolve_http_error_returns_none():
    # tests that an http error results in none
    test_url = "http://test.com/error"
    respx.get(test_url).mock(side_effect=httpx.RequestError("test error"))

    resolved = await resolve_magnet_uri(test_url)
    assert resolved is None


@respx.mock
async def test_resolve_unhandled_url_returns_none():
    # tests that a normal url (e.g. html page) returns none
    test_url = "http://test.com/page.html"
    respx.get(test_url).mock(httpx.Response(200, content=b"<html></html>"))

    resolved = await resolve_magnet_uri(test_url)
    assert resolved is None



from typing import Any, cast

import pytest
from textual.widgets import ListView

from torrra.app import TorrraApp
from torrra.core.config import Config
from torrra.screens.theme_selector import ThemeSelectorScreen


@pytest.fixture
def app(app_factory: Any) -> TorrraApp:
    app = app_factory()
    app.theme = "textual-dark"  # default theme
    return app


async def test_theme_selector_opens_and_closes_with_escape(app: TorrraApp):
    async with app.run_test() as pilot:
        await pilot.press("ctrl+t")
        assert isinstance(app.screen, ThemeSelectorScreen)
        assert len(app.screen_stack) == 3  # default + welcome + theme switcher

        await pilot.press("escape")
        assert len(app.screen_stack) == 2  # default + welcome screen


async def test_theme_selector_navigation_with_j_and_k(app: TorrraApp):
    async with app.run_test() as pilot:
        await pilot.press("ctrl+t")
        assert isinstance(app.screen, ThemeSelectorScreen)

        list_view = app.screen.query_one("ListView", ListView)
        initial_index = cast(int, list_view.index)

        await pilot.press("j")
        assert list_view.index == initial_index + 1

        await pilot.press("k")
        assert list_view.index == initial_index


@pytest.mark.usefixtures("fast_sleep")
async def test_theme_selector_select_theme_with_enter(
    app: TorrraApp, mock_config: Config, monkeypatch: pytest.MonkeyPatch
):
    # patch config instance used by theme_selector module
    # with mock_config
    monkeypatch.setattr(
        "torrra.screens.theme_selector.get_config",
        lambda: mock_config,
    )

    async with app.run_test() as pilot:
        await pilot.press("ctrl+t")
        assert isinstance(app.screen, ThemeSelectorScreen)

        list_view = app.screen.query_one("ListView", ListView)
        target_theme = list_view.children[1].name

        # navigate and wait for preview
        list_view.index = 1
        await pilot.pause()
        assert app.theme == target_theme

        await pilot.press("enter")
        await pilot.pause()

        assert len(app.screen_stack) == 2  # default + welcome screen
        assert app.theme == target_theme
        assert mock_config.get("general.theme") == target_theme


async def test_theme_selector_cancel_selection_with_escape(
    app: TorrraApp, mock_config: Config
):
    original_theme = app.theme
    mock_config.set("general.theme", original_theme)

    async with app.run_test() as pilot:
        await pilot.press("ctrl+t")
        assert isinstance(app.screen, ThemeSelectorScreen)

        list_view = app.screen.query_one("ListView", ListView)
        target_theme = list_view.children[1].name

        # navigate and wait for preview
        list_view.index = 1
        await pilot.pause(0.6)

        assert app.theme == target_theme

        await pilot.press("escape")
        await pilot.pause()

        assert len(app.screen_stack) == 2  # default + welcome screen
        assert app.theme == original_theme
        assert mock_config.get("general.theme") == original_theme



import subprocess
from typing import cast

from textual import on, work
from textual.app import ComposeResult
from textual.containers import Vertical
from textual.message import Message
from textual.widgets import Input, Static
from typing_extensions import override

from torrra._types import Indexer, Torrent
from torrra.core.config import get_config
from torrra.core.constants import DEFAULT_MAX_RETRIES, DEFAULT_TIMEOUT
from torrra.core.torrent import get_torrent_manager
from torrra.indexers.base import BaseIndexer
from torrra.utils.helpers import human_readable_size, lazy_import
from torrra.utils.magnet import resolve_magnet_uri
from torrra.widgets.data_table import AutoResizingDataTable
from torrra.widgets.details_panel import DetailsPanel
from torrra.widgets.spinner import Spinner


class SearchContent(Vertical):
    COLS: list[tuple[str, str, int]] = [
        ("No", "no_col", 2),
        ("Title", "title_col", 25),
        ("Size", "size_col", 10),
        ("S:L", "seeders_leechers_col", 6),
    ]

    class SearchResults(Message):
        def __init__(self, results: list[Torrent], query: str) -> None:
            self.results: list[Torrent] = results
            self.query: str = query
            super().__init__()

    class DownloadRequested(Message):
        def __init__(self, torrent: Torrent) -> None:
            self.torrent: Torrent = torrent
            super().__init__()

    def __init__(self, indexer: Indexer, search_query: str, use_cache: bool):
        super().__init__(id="search_content")
        self.indexer: Indexer = indexer
        self.search_query: str = search_query
        self.use_cache: bool = use_cache

        # instance-level cache
        self._indexer_instance_cache: BaseIndexer | None = None

        # application states
        self._search_results_map: dict[str, Torrent] = {}
        self._selected_torrent: Torrent | None = None

        # ui refs (cached later)
        self._search_input: Input
        self._table: AutoResizingDataTable[str]
        self._details_panel: DetailsPanel
        self._loader: Vertical

    @override
    def compose(self) -> ComposeResult:
        yield Input(placeholder="Search...", value=self.search_query)
        yield AutoResizingDataTable(cursor_type="row", classes="hidden")
        yield DetailsPanel()
        with Vertical(id="loader"):
            yield Static()
            yield Spinner(name="shark")

    def on_mount(self) -> None:
        self._search_input = self.query_one(Input)
        self._search_input.border_title = "search"
        self._search_input.focus()

        self._table = self.query_one(AutoResizingDataTable)
        self._table.expand_col = "title_col"
        self._table.border_title = "results"

        self._details_panel = self.query_one(DetailsPanel)
        self._details_panel.border_title = "details"

        self._loader = self.query_one("#loader", Vertical)
        # setup table
        for label, key, width in self.COLS:
            self._table.add_column(label, width=width, key=key)
        # send initial search
        self.post_message(Input.Submitted(self._search_input, self.search_query))

    async def key_enter(self) -> None:
        if not self._details_panel.has_focus:
            return

        if self._selected_torrent:
            # uri returned from the indexer
            # might not be a proper magnet uri, resolve anyways
            raw_magnet_uri = self._selected_torrent.magnet_uri
            resolved_magnet_uri = await resolve_magnet_uri(raw_magnet_uri)

            if resolved_magnet_uri is None:
                return

            # update with resolved magnet_uri
            self._selected_torrent.magnet_uri = resolved_magnet_uri

            config = get_config()
            if config.get("general.download_in_external_client", False):
                if config.get("general.use_transmission", False):
                    tran_user = config.get("general.transmission_user", "")
                    tran_pass = config.get("general.transmission_pass", "")

                    subprocess.run(
                        [
                            "transmission-remote",
                            "--auth",
                            tran_user + ":" + tran_pass,
                            "-a",
                            resolved_magnet_uri,
                        ],
                        capture_output=True,
                        text=True,
                    )
                    self.notify(
                        "Opened in [b]transmission-remote[/b]",
                        title="Torrent Opened",
                    )
                else:
                    self.app.open_url(resolved_magnet_uri)
                    self.notify(
                        "Opened in default magnet: handler",
                        title="Torrent Opened",
                    )
            else:  # continue with libtorrent
                tm = get_torrent_manager()
                tm.add_torrent(self._selected_torrent)
                title = self._selected_torrent.title
                short_title = (title[:30] + "...") if len(title) > 30 else title
                self.notify(
                    f"Started downloading [b]{short_title}[/b]",
                    title="Download Started",
                )
                self.post_message(self.DownloadRequested(self._selected_torrent))

    def on_input_submitted(self, event: Input.Submitted) -> None:
        query = event.value
        if not query:
            return

        self._table.add_class("hidden")
        self._table.clear()
        self._details_panel.add_class("hidden")
        self._loader.remove_class("hidden")
        cast(Spinner, self._loader.children[1]).resume()
        cast(Static, self._loader.children[0]).update(
            f"Searching for [b]{query}[/b]..."
        )

        self._perform_search(query)

    @work(exclusive=True)
    async def _perform_search(self, query: str) -> None:
        try:
            indexer = self._get_indexer_instance()
            results = await indexer.search(query, use_cache=self.use_cache)
            self.post_message(self.SearchResults(results or [], query))
        except Exception:
            self.notify(
                "Search failed, check indexer settings",
                title="Search Failed",
                severity="error",
            )  # post empty results just to stop spinners
            self.post_message(self.SearchResults([], query))

    @on(SearchResults)
    def on_search_results(self, message: SearchResults) -> None:
        if not message.results:
            cast(Spinner, self._loader.children[1]).pause()
            cast(Static, self._loader.children[0]).update(
                f"Nothing Found for [b]{message.query}[/b]"
            )  # show loader and exit
            return

        self._loader.add_class("hidden")
        self._table.remove_class("hidden")
        self._table.focus()  # initial focus table
        self._table.border_title = f"results ({len(message.results)})"

        seen: set[str] = set()
        for idx, torrent in enumerate(message.results):
            if torrent.magnet_uri in seen:
                continue

            seen.add(torrent.magnet_uri)
            self._search_results_map[torrent.magnet_uri] = torrent
            self._table.add_row(
                str(idx + 1),
                torrent.title,
                human_readable_size(torrent.size),
                f"{str(torrent.seeders)}:{str(torrent.leechers)}",
                key=torrent.magnet_uri,
            )

    def on_details_panel_closed(self):
        self._selected_torrent = None
        self._table.focus()

    def on_data_table_row_selected(
        self, event: AutoResizingDataTable.RowSelected
    ) -> None:
        magnet_uri = cast(str, event.row_key.value)
        self._selected_torrent = self._search_results_map.get(magnet_uri)
        if not self._selected_torrent:
            return

        details = f"""
[b]{self._selected_torrent.title}[/b]
[b]Size:[/b] {human_readable_size(self._selected_torrent.size)} - [b]Seeders:[/b] {self._selected_torrent.seeders} - [b]Leechers:[/b] {self._selected_torrent.leechers} - [b]Source:[/b] {self._selected_torrent.source}

[dim]Press 'enter' to download or 'esc' to close.[/dim]
"""

        self._details_panel.update_content(details.strip())
        self._details_panel.remove_class("hidden")
        self._details_panel.focus()

    def _get_indexer_instance(self) -> BaseIndexer:
        if self._indexer_instance_cache:
            return self._indexer_instance_cache

        name = self.indexer.name
        indexer_cls_str = f"torrra.indexers.{name}.{name.title()}Indexer"

        indexer_cls = lazy_import(indexer_cls_str)
        assert issubclass(indexer_cls, BaseIndexer)
        indexer_instance = indexer_cls(
            url=self.indexer.url,
            api_key=self.indexer.api_key,
            timeout=get_config().get("general.timeout", DEFAULT_TIMEOUT),
            max_retries=get_config().get("general.max_retries", DEFAULT_MAX_RETRIES),
        )

        self._indexer_instance_cache = indexer_instance
        return indexer_instance



from typing import Any
from unittest.mock import MagicMock

from textual.pilot import Pilot

from torrra._types import Torrent


def test_home_screen_snapshot(
    app_factory: Any, mock_indexer: MagicMock, snap_compare: Any
):
    # return mock torrents as result
    mock_indexer.search.return_value = [
        Torrent(
            magnet_uri="magnet:?xt=urn:btih:arch_new",
            title="Arch Linux 2025.11.01",
            size=1073741824,
            seeders=850,
            leechers=50,
            source="LinuxTacker",
        ),
        Torrent(
            magnet_uri="magnet:?xt=urn:btih:arch_old",
            title="Arch Linux 2024.01.01",
            size=838860800,
            seeders=5,
            leechers=15,
            source="LinuxTacker",
        ),
    ]

    async def run_before(pilot: Pilot[Any]):
        await pilot.pause()

    app = app_factory("arch linux iso")
    app.theme = "textual-dark"  # default theme
    assert snap_compare(app, run_before=run_before)


def test_welcome_screen_snapshot(app_factory: Any, snap_compare: Any):
    async def run_before(pilot: Pilot[Any]):
        await pilot.press(*list("arch linux iso"))
        await pilot.pause()

    app = app_factory()
    app.theme = "textual-dark"  # default theme
    assert snap_compare(app, run_before=run_before)


def test_theme_selector_screen_snapshot(app_factory: Any, snap_compare: Any):
    async def run_before(pilot: Pilot[Any]):
        await pilot.press("ctrl+t")
        await pilot.pause()

    app = app_factory()
    app.theme = "textual-dark"  # default theme
    assert snap_compare(app, run_before=run_before)



from unittest.mock import MagicMock

import pytest
from click.testing import CliRunner

from torrra.__main__ import cli
from torrra._version import __version__


def test_cli_version():
    # tests that the --version flag works
    runner = CliRunner()
    result = runner.invoke(cli, ["--version"])
    assert result.exit_code == 0
    assert __version__ in result.output


def test_search_command_calls_runner(monkeypatch: pytest.MonkeyPatch):
    # tests that the 'search' command calls the correct underlying function
    mock_run_func = MagicMock()
    monkeypatch.setattr("torrra.utils.indexer.run_with_default_indexer", mock_run_func)

    runner = CliRunner()
    result = runner.invoke(cli, ["search", "arch linux iso", "--no-cache"])

    assert result.exit_code == 0
    # assert that the function was called once with the correct arguments
    mock_run_func.assert_called_once_with(no_cache=True, search_query="arch linux iso")


def test_prowlarr_command_calls_runner_with_cache(monkeypatch: pytest.MonkeyPatch):
    # tests that the "prowlarr" command calls the correct underlying function
    mock_run_func = MagicMock()
    monkeypatch.setattr("torrra.utils.indexer.run_with_indexer", mock_run_func)

    runner = CliRunner()
    result = runner.invoke(
        cli,
        [
            "prowlarr",
            "--url",
            "http://mock.indexer.url",
            "--api-key",
            "mock_api_key",
            "--no-cache",
        ],
    )

    assert result.exit_code == 0
    mock_run_func.assert_called_once_with(
        name="prowlarr",
        indexer_cls_str="torrra.indexers.prowlarr.ProwlarrIndexer",
        url="http://mock.indexer.url",
        api_key="mock_api_key",
        no_cache=True,
    )


@pytest.mark.usefixtures("mock_config")
def test_config_commands_flow():
    # tests the full get/set/list flow for the config command
    runner = CliRunner()

    # test setting a value
    set_result = runner.invoke(cli, ["config", "set", "test.key", "test_value"])
    assert set_result.exit_code == 0

    # test getting the value back
    get_result = runner.invoke(cli, ["config", "get", "test.key"])
    assert get_result.exit_code == 0
    assert "test_value" in get_result.output

    # test listing all values
    list_result = runner.invoke(cli, ["config", "list"])
    assert list_result.exit_code == 0
    assert "test.key=test_value" in list_result.output



import pytest

from torrra.core import config as config_module
from torrra.core.config import Config
from torrra.core.exceptions import ConfigError


def test_config_initialization_creates_default_file(mock_config: Config):
    # test that a default config file is created
    config_path = config_module.CONFIG_FILE
    assert config_path.exists()
    # check for a known default value
    assert mock_config.get("general.theme") == "textual-dark"


def test_config_get_existing_value(mock_config: Config):
    # test getting a pre-existing value
    assert mock_config.get("general.use_cache") is True


def test_config_get_nonexistent_with_default(mock_config: Config):
    # test that a default value is returned for a non-existent key
    assert mock_config.get("general.non_existent_key", "default") == "default"


def test_config_get_nonexistent_raises_error(mock_config: Config):
    # test that getting a non-existent key w/o a default raises ConfigError
    with pytest.raises(ConfigError, match="key does not contain a section"):
        mock_config.get("general.non_existent_key")


def test_config_get_section_raises_error(mock_config: Config):
    # test that trying to get a value from a section key raises ConfigError
    with pytest.raises(ConfigError, match="key does not contain a value"):
        mock_config.get("general")


def test_config_set_and_get_new_value(mock_config: Config):
    # test setting a new value and then getting it back
    mock_config.set("new.section.key", "new_value")
    assert mock_config.get("new.section.key") == "new_value"


def test_config_set_type_conversion(mock_config: Config):
    # test that string values are correctly converted to other types
    mock_config.set("types.bool_true", "true")
    mock_config.set("types.bool_false", "False")
    mock_config.set("types.integer", "123")
    mock_config.set("types.float", "45.6")

    assert mock_config.get("types.bool_true") is True
    assert mock_config.get("types.bool_false") is False
    assert mock_config.get("types.integer") == 123
    assert mock_config.get("types.float") == 45.6


def test_config_list_flattens_correctly(mock_config: Config):
    # test that list method correctly flattens
    mock_config.set("general.theme", "new-theme")
    mock_config.set("new.section.key", "value")
    mock_config.set("new.section.bool", "true")

    config_list = mock_config.list()
    # the order can vary, so we check for presence instead of exact list match
    assert "general.theme=new-theme" in config_list
    assert "new.section.key=value" in config_list
    assert "new.section.bool=true" in config_list
    # check a default value is also present
    assert "general.download_in_external_client=false" in config_list



from typing import cast
from unittest.mock import MagicMock

import pytest
from textual.coordinate import Coordinate
from textual.widgets import DataTable, Static

from torrra._types import Indexer, Torrent
from torrra.app import TorrraApp
from torrra.screens.home import HomeScreen


@pytest.fixture
def app():
    # start app on SearchScreen
    # by providing a default search_query
    return TorrraApp(
        indexer=Indexer(
            name="jackett", url="http://mock.indexer.url", api_key="mock_api_key"
        ),
        use_cache=False,
        search_query="arch linux iso",
    )


async def test_home_screen_search(app: TorrraApp, mock_indexer: MagicMock):
    mock_indexer.search.return_value = [
        Torrent(
            magnet_uri="magnet:?xt=urn:btih:mock",
            title="Arch Linux ISO (Mock)",
            size=840499200,
            seeders=523,
            leechers=17,
            source="MockIndexer",
        )
    ]

    async with app.run_test():
        assert isinstance(app.screen, HomeScreen)

        table = cast(
            DataTable[str], app.screen.query_one("SearchContent DataTable", DataTable)
        )
        # table should have results
        assert not table.has_class("hidden")

        # verify table has the correct number of rows and content
        assert table.row_count == len(mock_indexer.search.return_value)
        assert table.get_cell_at(Coordinate(0, 1)) == "Arch Linux ISO (Mock)"


async def test_home_screen_search_no_results(app: TorrraApp, mock_indexer: MagicMock):
    # ensure result is empty []
    mock_indexer.search.return_value = []

    async with app.run_test():
        assert isinstance(app.screen, HomeScreen)

        loader_status = app.screen.query_one("#loader Static", Static)
        table = app.screen.query_one("SearchContent DataTable")

        assert "Nothing Found" in str(loader_status.content)
        assert table.has_class("hidden")
